(**************************************************************************

   	This file is part of the AtomicDensityMatrix package for Mathematica
    Copyright (C) 2004-2017 Simon M. Rochester <simon@rochesterscientific.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

**************************************************************************)


BeginPackage["AtomicDensityMatrix`Observables`", {
    "AtomicDensityMatrix`",
    "AtomicDensityMatrix`Common`",
    "AtomicDensityMatrix`CouplingCoefficients`",
    "AtomicDensityMatrix`TensorAlgebra`",
    "AtomicDensityMatrix`AtomicSystem`",
    "AtomicDensityMatrix`ExperimentalSystem`",
    "AtomicDensityMatrix`DensityMatrix`",
    "AtomicDensityMatrix`OpticalField`",
    "AtomicDensityMatrix`RotatingWaveApproximation`",
    "AtomicDensityMatrix`WignerEckart`",
    "AtomicDensityMatrix`WignerRotate`"
}];

Begin["`Private`"];

WaveEqnSolution[w_, "AngleEllipticity", {{E0_, cphi_}, {al_, ep_}}, {P1_, P2_, P3_, P4_}] :=
    {{((2 * Pi * w) / E0) * (Sin[al] * (Sin[ep] * ((-P1) * Cos[cphi] + P2 * Sin[cphi]) +
        Cos[ep] * (P4 * Cos[cphi] + P3 * Sin[cphi])) + Cos[al] * (Cos[ep] * (P2 * Cos[cphi] + P1 * Sin[cphi]) +
        Sin[ep] * (P3 * Cos[cphi] - P4 * Sin[cphi]))),
        If[Abs@N@ep === Pi / 4., Indeterminate,
            ((2 * Pi * w) / E0) * Sec[2 * ep] *
                (Cos[al] * (Cos[ep] * (P1 * Cos[cphi] - P2 * Sin[cphi]) + Sin[ep] * (P4 * Cos[cphi] + P3 * Sin[cphi])) -
                    Sin[al] * (Sin[ep] * (P2 * Cos[cphi] + P1 * Sin[cphi]) + Cos[ep] * ((-P3) * Cos[cphi] + P4 * Sin[cphi])))]
    },
        {If[Abs@N@ep === Pi / 4., Indeterminate,
            ((2 * Pi * w) / E0) * Sec[2 * ep] * (Cos[al] * (Sin[ep] * (P1 * Cos[cphi] - P2 * Sin[cphi]) +
                Cos[ep] * (P4 * Cos[cphi] + P3 * Sin[cphi])) - Sin[al] * (Cos[ep] * (P2 * Cos[cphi] + P1 * Sin[cphi]) +
                Sin[ep] * ((-P3) * Cos[cphi] + P4 * Sin[cphi])))],
            (-((2 * Pi * w) / E0)) *
                (Sin[al] * (Cos[ep] * (P1 * Cos[cphi] - P2 * Sin[cphi]) + Sin[ep] * (P4 * Cos[cphi] + P3 * Sin[cphi])) +
                    Cos[al] * (Sin[ep] * (P2 * Cos[cphi] + P1 * Sin[cphi]) + Cos[ep] * ((-P3) * Cos[cphi] + P4 * Sin[cphi])))}};

WaveEqnSolution[w_, "RatioPhase", {{E0_, cphi_}, {chi_, ph_}}, {P1_, P2_, P3_, P4_}] :=
    {{((2 * Pi * w) / E0) * (Cos[cphi] * (P2 * Cos[chi] + (P4 * Cos[ph] + P3 * Sin[ph]) * Sin[chi]) +
        Sin[cphi] * (P1 * Cos[chi] + (P3 * Cos[ph] - P4 * Sin[ph]) * Sin[chi])),
        (2 * Pi * w * Sec[chi] * (P1 * Cos[cphi] - P2 * Sin[cphi])) / E0},
        {2 * Pi * w * ((P4 * Cos[ph + cphi] * Cos[chi] + P3 * Cos[chi] * Sin[ph + cphi] -
            (P2 * Cos[cphi] + P1 * Sin[cphi]) * Sin[chi]) / E0), If[chi === 0, 0,
            (1 / E0) * 2 * Pi * w * (P3 * Cos[ph + cphi] * Csc[chi] - P1 * Cos[cphi] * Sec[chi] + P2 * Sec[chi] * Sin[cphi] -
                P4 * Csc[chi] * Sin[ph + cphi])]}};

WaveEqnSolution[w_, "Stokes", {{E0_, cphi_}, {S1_, S2_, S3_}}, {P1_, P2_, P3_, P4_}] :=
    {{((Sqrt[2] * Pi * w) / (E0 * Sqrt[1 + S1])) * (P2 * (1 + S1) * Cos[cphi] + P1 * (1 + S1) * Sin[cphi] +
        Sqrt[1 - S1^2] * (P4 * Cos[cphi + ArcTan[S2, S3]] + P3 * Sin[cphi + ArcTan[S2, S3]])),
        (2 * Sqrt[2] * Pi * w * (P1 * Cos[cphi] - P2 * Sin[cphi])) / (E0 * Sqrt[1 + S1])},
        {(-((2 * Sqrt[2] * Pi * w) / E0)) * (P2 * (-1 + S1) * Sqrt[1 + S1] * Cos[cphi] + P4 * Sqrt[1 - S1] * (1 + S1) *
            Cos[cphi + ArcTan[S2, S3]] + P1 * (-1 + S1) * Sqrt[1 + S1] * Sin[cphi] +
            P3 * Sqrt[1 - S1] * (1 + S1) * Sin[cphi + ArcTan[S2, S3]]),
            (-((2 * Sqrt[2] * Pi * w) / (E0 * (-1 + S1) * Sqrt[1 + S1]))) * ((-1 + S1) * (P2 * S1 * S2 - P1 * S3) * Cos[cphi] +
                Sqrt[1 - S1^2] * (P4 * S1 * S2 - P3 * S3) * Cos[cphi + ArcTan[S2, S3]] + (-1 + S1) * (P1 * S1 * S2 + P2 * S3) *
                Sin[cphi] + Sqrt[1 - S1^2] * (P3 * S1 * S2 + P4 * S3) * Sin[cphi + ArcTan[S2, S3]]),
            (-((2 * Sqrt[2] * Pi * w) / (E0 * (-1 + S1) * Sqrt[1 + S1]))) * ((-1 + S1) * (P1 * S2 + P2 * S1 * S3) * Cos[cphi] +
                Sqrt[1 - S1^2] * (P3 * S2 + P4 * S1 * S3) * Cos[cphi + ArcTan[S2, S3]] + (-1 + S1) * ((-P2) * S2 + P1 * S1 * S3) *
                Sin[cphi] + Sqrt[1 - S1^2] * ((-P4) * S2 + P3 * S1 * S3) * Sin[cphi + ArcTan[S2, S3]])}};

(*Lab frame density matrix in terms of the rotating-frame density-matrix elements. Under a passive transformation U, we have  |*)
(*Overscript[\[Psi], ~]\[RightAngleBracket]=U^\[Dagger]|\[Psi]\[RightAngleBracket]. Thus *)
(*Overscript[\[Rho], ~]=|*)
(*Overscript[\[Psi], ~]\[RightAngleBracket]\[LeftAngleBracket]*)
(*Overscript[\[Psi], ~]|=U^\[Dagger]|\[Psi]\[RightAngleBracket]\[LeftAngleBracket]\[Psi]|U=U^\[Dagger] \[Rho] U, so \[Rho]=U *)
(*Overscript[\[Rho], ~](U^\[Dagger]).*)

DMFromRotatingDM[system : {_AtomicState..}, RWT_?MatrixQ, opts___Rule] :=
    Module[{newopts = FilterRules[DeleteCases[{opts}, _ -> Automatic], Options[DensityMatrix]]},
        RWT.DensityMatrix[system, Sequence @@ newopts].ECT[RWT]
    ];

Polarization[system : {_AtomicState..}, RWT_, opts___Rule] :=
    Tr[DMFromRotatingDM[system, RWT, opts].#]& /@ ToCartesian[WignerEckart[system, {Dipole, 1}]];

Options[PolarizationComponents] = Join[
    DeleteCases[Options[OpticalField], CartesianCoordinates -> _],
    Options[DensityMatrix],
    {TransformMatrix -> Automatic}
] /. (x_ -> _) -> (x -> Automatic);

(*Not sure why I enforced complex expanding the variables here. Do I have to?*)

PolarizationComponents[system : {_AtomicState..}, args__, opts : OptionsPattern[]] /; 2 <= Length[{args}] <= 3 :=
    Module[ {w, k0, E0, cphi, pol, polparams, e, k, rwt, parameterization,
        newopts = Append[DeleteCases[Join[{opts}, Options[DensityMatrix]], _ -> Automatic] /.
            (ComplexExpandVariables -> False) -> (ComplexExpandVariables -> True), ComplexExpandVariables -> True]},
        {{w, k0}, {E0, cphi}, polparams, {e, k}, parameterization} = standardizeField[{args}, {opts}, Options[PolarizationComponents]];
        rwt = If[OptionValue[TransformMatrix] === Automatic,
            RotatingWaveTransformMatrix[system, w],
            OptionValue[TransformMatrix]
        ];
        pol = Polarization[system, rwt, Sequence @@ newopts];
        If[ Length[pol] === 1, pol = pol[[1]]e ];
        Flatten[{1, -1}Coefficient[Expand@ExpToTrig[pol\[CenterDot]#], {Cos[Expand[w t]], Sin[Expand[w t]]}]& /@ {e, Cross[k, e]}]
    ];

Options[Observables] = Join[
    Join[
        DeleteCases[Options[OpticalField], CartesianCoordinates -> _],
        Options[DensityMatrix]
    ] /. (x_ -> _) -> (x -> Automatic),
    {ExpandDipoleRME -> True, TransformMatrix -> Automatic}
];

Observables[system : {_AtomicState..}, args__, opts : OptionsPattern[]] /; 2 <= Length[{args}] <= 3 :=
    Module[ {w, E0, cphi, pol, e, k, k0, newopts, wes, parameterization},
        {{w, k0}, {E0, cphi}, pol, {e, k}, parameterization} = standardizeField[{args}, {opts}, Options[Observables]];
        newopts = FilterRules[DeleteCases[Join[{opts}, Options[Observables]], _ -> Automatic], Options[PolarizationComponents]];
        wes = WaveEqnSolution[w, parameterization, {{E0, cphi}, pol}, PolarizationComponents[system, args, Sequence @@ newopts]];
        If[ OptionValue[ExpandDipoleRME],
            ExpandDipoleRME[system, wes],
            wes
        ]
    ];

Observables[as : {_AtomicState..}, es : {_ExperimentalRegion..}, opts : OptionsPattern[]] :=
    Total[Observables[as, Sequence @@ (OpticalParameters[#]), DMLabel -> RegionLabel[#], opts]& /@ es];

Voigt[u_, v_, 0] := v / (\[Pi] (u^2 + v^2));

Voigt[u_, 0, GD_] := 1 / (Sqrt[\[Pi]] GD) E^(-(u^2 / GD^2));

Voigt[u_, v_, GD_] := 1 / (GD Sqrt[\[Pi] ]) Re[E^-((u + I v) / GD)^2 Erfc[-I((u + I v) / GD)]];

InternalEnergy[s_] := WignerEckart[s, {Energy, 0, 0}, s];

Options[LinearAbsorption] = {ExpandDipoleRME -> True};

LinearAbsorption[s1_AtomicState, s2_AtomicState, GD_, w_, dz_, OptionsPattern[]] :=
    If[OptionValue[ExpandDipoleRME],
        ExpandDipoleRME[{s1, s2}, #], #]&[
        -((4 \[Pi]^2 (Energy[s2] - Energy[s1])) / SublevelMultiplicity[DeleteParameter[s1, F]]) *
            Voigt[ w - (InternalEnergy[s2] - InternalEnergy[s1]), NaturalWidth[s2] / 2, GD] dz^2
    ];

LinearAbsorption[sys : {_AtomicState..}, GD_, w_] := Module[{dz},
    dz = WignerEckart[sys, {Dipole, 1, 0}];
    Total[Outer[
        If[M[#1] === M[#2] || AtomicSystemType[sys] === "Toy",
            LinearAbsorption[#1, #2, GD, w, Component[sys, dz, {#1, #2}]],
            0
        ]&,
        GroundStates[sys],
        ExcitedStates[sys]
    ], 2]
];

End[];

EndPackage[];
